 export CIRCLE_BUILD_NUM=latest
 tests/scripts/build_docker_image.sh
 tests/scripts/decompress_test_data.sh
 tests/scripts/check_formatting.sh
 tests/scripts/python_tests.sh
 tests/scripts/tiff_cubing.sh
 tests/scripts/tile_cubing.sh
 tests/scripts/simple_tiff_cubing.sh
 tests/scripts/simple_tiff_cubing_no_compression.sh
 tests/scripts/knossos_conversion.sh
 tests/scripts/decompress_reference_mag.sh
 tests/scripts/downsampling.sh
 tests/scripts/compression.sh
 tests/scripts/in_place_compression.sh
 tests/scripts/meta_generation.sh
 tests/scripts/simple_anisotropic_tiff_cubing.sh